// HID serial transfer, suitable for low-speed devices.
// Unluckily, no terminal emulator supports HID.
// Here is no OUT endpoint/report so receiving data from host
// (i.e. keyboard, cursor position report) is not implemented here.
// Maximum transfer rate: 32 kByte/s

#include "usb.h"
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

/* Erklärung:
Das praktische Ziel ist ein USB-Gerät, welches treiberlos als HID-Gerät
(ab Windows 98), als WebUSB-Gerät (ab Windows 8.1) und als serielle
Schnittstelle (im USB-Sprech „CDC“ = Communication Device Class,
genauer „CDC-ACM“) funktioniert.

Dabei macht USB-CDC die größten Probleme:
- Solo funktioniert USB-CDC bereits ab Windows 98, erfordert jedoch
  stets eine INF-Datei sowie bis hin zu Windows 2000 einen Treiber (.sys).
  Man kennt dies von älteren USB-RS232-Adaptern.
  * Für Windows 98, Me und 2k musste der Adapterhersteller selber einen
    Treiber schreiben oder einen von Microsoft kaufen
  * Im Kaufvertrag stand ein Passus, der es verbietet, den Treiber
    generisch mit /allen/ USB-CDC-Adaptern (USB\Class_02)
    zu verbinden; die Verbindung /musste/ via eineindeutiger
    Vendor-ID + Product-ID erfolgen. Das maximierte Microsofts Absatz.
  * Ab Windows XP war usbser.sys als Treiber bereits im System vorhanden.
    Hersteller von USB-Seriell-Adaptern /mussten/ (laut Vertrag)
    ihren eigenen Treiber zurückziehen sowie die INF-Datei für
    64 Bit anpassen: Gesonderte Sections sind da erforderlich.
  * (Nicht nur) ich habe schließlich irgendeine dieser INF-Dateien
    hergenommen und generisch gemacht. Damit lassen sich USB-Seriell-Adapter
    nahezu jedes Herstellers sogar unter Windows 98 ohne lästiges
    Installationsgewurstel verwenden.
  * COM-Portnummern wechseln generell mit der Buchse.
    Hat das USB-Gerät eine Seriennummer (Device Descriptor Offset 16),
    merkt sich Windows (ab 2k) die Beziehung zwischen Nummer und jener
    Zeichenkette. Daher kommt es, dass einige CDC-Geräte ihre
    COM-Portnummer beim Umstöpseln behalten, die meisten jedoch nicht.

- In Kombination mit anderen USB-Funkionen ist USB-CDC nach USB-Standard
  anfangs nicht so recht vorgesehen. Denn CDC benötigt /zwei/ USB-
  Interfaces und kann so nicht Teil eines Muktifunktions-USB-Gerätes
  (USB\Class_00) sein: Windows lädt für die Geräteklasse 00 seinen
  Multifunktionstreiber (usbccgp.sys), der wiederum einen USB-Upstream-Port
  emuliert und für jedes enthaltene USB-Interface einzeln den passenden
  Treiber lädt. Daraufhin wird zwar usbser.sys für das entsprechende
  Control-Interface gefunden und geladen; für das Data-Interface wird
  jedoch nur ein Stub-Treiber geladen, der den sattsam bekannten
  Gerätetreiber-Installieren-Dialog hervorbringt.
  Und usbser.sys verweigert die Arbeit, weil es kein Data-Interface findet
  oder die Arbeit unter der Fuchtel von usbccgp.sys verweigert.
  Die (angeblich) richtige Lösung hierfür sei eine maßgeschneiderte
  INF-Datei, da habe ich aber nichts greifbares gefunden.

- Mit Windows 8.1, bzw. um diese Zeit wurde der USB-Standard 2.0 um einen
  „Interface Association Descriptor“ (IAD) erweitert.
  Damit soll genau dieses Problem angegangen werden:
  Der IAD gruppiert mehrere Interfaces zu dem Zweck, dass usbccgp.sys
  die Interfaces nicht zerteilt sondern als Bundle behält und dieses
  dem nachfolgenden Treiber (hier: usbser.sys) übergibt.
  Klar, usbccgp.sys hätte auch selber so schlau sein können,
  die Klasse 02 (CDC sowie CDC-Control) und die Klasse 0A (CDC-Data)
  zusammen lassen zu können. Dafür gibt es sogar der Klasse 02
  untergeordnete Deskriptoren. Hat aber Microsoft wegignoriert und lieber
  im USB-Konsortium insistiert: Will /übergeordneten/ Desktiptor.
  Bei Linux ist man genau jenen Weg gegangen, und auch für eingebettete
  Systeme ist das der gängige Weg.
  Der Pferdefuß ist nun, dass der IAD erfordert, dass das USB-Gerät
  sich mit USB-Standard 2.0 (statt wie sonst 1.1) meldet
  und dass die USB-Geräteklasse nicht mehr 00 sondern EF sein muss.
  Damit erscheint bei Mutifunktionsgeräten mit HID und CDC und IAD
  alles problemlos unter Windows XP SP3+, jedoch funktionert darunter
  /gar nichts/ mehr: Denn auch das HID-Interface verschwindet,
  da bspw. Windows98 mit der neuen Geräteklasse EF nichts anfangen kann.
  Kann man zwar beheben, indem man eine INF-Datei beilegt, die
  die USB-Geräteklasse EF mit usbccgp.sys verbindet, aber das verfehlt
  erstmalig das Ziel, mit HID einen treiberlosen Zugang für /jede/
  Windows-Version bereit zu stellen.

Die Moral von der Geschicht: Entweder man benutze CDC solo, oder man
verzichte darauf. Jede Art von Kombination schränkt die Portabilität ein.
Wenn schon kombiniert, dann allenfalls als Debug-Interface, nicht für das
Fertigprodukt.

Für Mikrocontroller ohne USB-Hardware, typischwerweise mit obdev.at V-USB,
ist CDC nur bis Windows XP verwendbar.
Höhere Windows-Versionen verbieten das, was auch per USB-Standard verboten
ist, zumal CDC auf Low-Speed-USB eine hohe Belastung für den Bus
und den Mikrocontroller darstellt: Jener wird nahezu lückenlos mit
Datenlese-Tokens bombardiert, die er mit NAK beantworten muss.
(Trifft auf den hier verwendeten AT90USB82 und Nachfolger nicht zu.)

Was wäre die /richtige/ Lösung gewesen?
Richtig gewesen wäre eine Neustrukturierung der USB-CDC-Geräteklasse
sowie aller anderen USB-Klassen, die mehrere Interfaces belegen
(USB-Audio, USB-Video), zu einer 1:1-Beziehung, also Integration der
Dateninterfaces in das jeweilige Steuer-Interface.
Wurde bis heute nicht darüber nachgedacht.
Aber USB ist wie Unicode ein Politikum: Es wird nur 'rangeflickt, nichts
aufgeräumt. Nur dass das bei Unicode von vornherein angesagtes Ziel ist.
Beim USB-Standard habe ich hingegen keinen derartigen Passus gelesen.

Die /zweite richtige Lösung/ wäre ein HID-Substandard für eine
Terminal-Emulation.
Und ein PuTTY, welches derartige HID-Geräte ansprechen kann.
Zwar mit vermindertem Durchsatz gegenüber CDC, aber ohne den ganzen Ärger:
* Zwang zu USB Full-Speed
* Wechselnde COM-Portnummern
* Oben angeführte Kompatibilitätsprobleme in Multifunktionsgeräten


*WebUSB* ist ebenfalls ein heißes Eisen, verträgt sich allerdings
von vornherein mit einem benachbarten HID-Interface und stellt so
maximale Portabilität sicher.
Ein WebUSB-Interface ist der Schlüssel für den Zugriff auf ein USB-Gerät
mittels Smartphone oder Smart-TV. Das kann bei Bastelprojekten das lästige
Design eines Benutzerinterfaces in Hardware ersparen oder reduzieren.
Als Beispiel sei ein Rasenmähroboter genannt, der zur Lokalbedienung
nur 3 Knöpfe (Stop, Home, Work) und wenige Anzeigen
(Homing, Working, Collision+Lift Detect, Battery Indicator) benötigt.
Alles andere (Uhrzeit, Kalender, Passwort, dGPS-Setup wenn vorhanden)
dann mittels Smartphone.

Zu beachten ist, dass Firefox seine Zusammenarbeit mit WebUSB bis heute
verweigert. Alle anderen Browser können das mittlerweile.
Der erste war Google Chrome, dicht gefolgt von Microsoft Edge.
WebUSB funktioniert etwa ab Android 7 auf Galaxy A5 (2016).


 *  Endpoint Buffer Configuration
Der AT90USB162 hat nur wenig Endpoint-RAM (176 Bytes)
und nur wenig (fünf) Endpoints, davon nur zwei mit Doppelpufferung.
Trotzdem reicht das gut für ein HID+CDC-Kombidevice oder ein Audio-Device.
Als sinnvolle Endpoint-Größe erscheint:
Anwendung			EP0	EP1	EP2	EP3	EP4	Dummy
CDC, symmetrisch		32	 -	16i	32i×2	32o×2	 -
CDC, asymmetrisch		 8	 -	 8i	64i×2	16o×2	 -
HID+CDC, symmetrisch		16	16i	16i	32i×2	32o×2	 -
HID+CDC, asymmetrisch		32	32i	16i	32i×2	16o×2	 -
HID+CDC, stark asymmetrisch	 8	16i	 8i	64i×2	 8o×2	 -
WebUSB+CDC (Oszilloskop)	16	16i	16o	32i×2	32i×2	 8i
WebUSB+HID+CDC (Oszilloskop)	16	16i	16i	32i×2	32i×2	 8i
WebUSB+HID+CDC (NC-Maschine)	16	16i	16o	32i×2	32o×2	 8i
HiFi-Stereo-Mikrofon		32	16i	 -	64i×2?	 -	 -
HiFi-Lautsprecher/Kopfhörer	32	16i	 -	 -	64o×2?	 -
Headset, Telefon		32	16i	 -	32i×2	32o×2	 -
Kamera, Massenspeicher		Nicht mit USB Full-Speed empfehlenswert
 **************************************************************************/
namespace usb{
/**************************************************************************
 *  Descriptor Data
 **************************************************************************/

// Descriptors are the data that your computer reads when it auto-detects
// this USB device (called "enumeration" in USB lingo).  The most commonly
// changed items are editable at the top of this file.  Changing things
// in here should only be done by those who've read chapter 9 of the USB
// spec and relevant portions of any USB class specifications!

#define W(x) (byte)(x),(byte)((x)>>8)

static const byte PROGMEM DeviceDesc[] = {
 18,		// bLength
 1,		// bDescriptorType
#if WUB
 W(0x0210),	// bcdUSB		2.1
#elif CDC*IAD
 W(0x0200),	// bcdUSB		2.0
#else
 W(0x0110),	// bcdUSB		1.1
#endif
#if CDC		// Class,SubClass,Protocol 
# if IAD
 239,2,1,	// misc. device		Not understood by Windows < XP SP3
# else
 2,2,1,		// communication device
# endif
#else
 0,0,0,		// see interface descriptor(s)
#endif
 1<<EP0SSH,	// bMaxPacketSize0
 W(9025/*0x03EB*/),	// idVendor		Atmel:1003
 W(0x2FF5),	// idProduct		?
 DATE_M,DATE_Y,	// bcdDevice		1.0
 1,		// iManufacturer
 2,		// iProduct
 3,		// iSerialNumber	So Windows COM port won't move
 1,		// bNumConfig
};

#if HID
static const byte PROGMEM HidRepDesc[] = {
 0x06, W(0xFF31),	// Usage Page 0xFF31 (vendor defined)
 0x09, 0x74,		// Usage 0x74
 0xA1, 0x53,		// Collection 0x53
  0x75, 8,		// report size = 8 bits
  0x15, 0,		// logical minimum = 0
  0x26, W(255),		// logical maximum = 255
  0x95, 1<<EP1SSH,	// report count: 16 byte
  0x09, 0x75,		// usage
  0x82, W(0x102),	// In (var,bufbytes)
  0x09, 0x76,		// usage
  0x92, W(0x102),	// Out (var,bufbytes)
  0x95, 8,		// report count: 8 byte
  0x09, 0x77,		// usage
  0xB1, 2,		// Feature (var)
 0xC0			// end collection
};
# define HID_DESC_OFFSET  (9+9)
#endif

static const byte PROGMEM ConfigDesc[9 + (9+9+7)*HID + (9+7*!(HID&CDC)+7)*WUB
  + (8*IAD+9+5+4+5+5+7+9+7+7)*CDC] = {
// configuration descriptor, USB spec 9.6.3, page 264-266, Table 9-10
 9, 			// bLength
 2,			// bDescriptorType	configuration
 W(sizeof ConfigDesc),	// wTotalLength
 HID+WUB+2*CDC,		// bNumInterfaces	HID + CDC Control + CDC Data
 1,			// bConfigurationValue
 0,			// iConfiguration	no string descriptor
 0x80,			// bmAttributes		Bit6: SelfPower
 25,			// bMaxPower		50 mA
#if HID
// interface descriptor, USB spec 9.6.5, page 267-269, Table 9-12
 9,			// bLength
 4,			// bDescriptorType
 0,			// bInterfaceNumber
 0,			// bAlternateSetting
 1,			// bNumEndpoints
 3,			// bIfaceClass		HID
 0,			// bIfaceSubClass	no boot support
 0,			// bIfaceProtocol
 4*(WUB|CDC),		// iInterface
// HID interface descriptor, HID 1.11 spec, section 6.2.1
 9,			// bLength
 0x21,			// bDescriptorType
 W(0x0110),		// bcdHID
 0,			// bCountryCode
 1,			// bNumDescriptors
 0x22,			// bDescriptorType	Report
 W(sizeof(HidRepDesc)),	// wDescriptorLength
// endpoint descriptor, USB spec 9.6.6, page 269-271, Table 9-13
// Während CDC ganz gut mit einem toten Interrupt-Endpoint zurechtkommt,
// führt das bei HID zum Geräteversagen.
 7,			// bLength
 5,			// bDescriptorType
 0x81,			// bEndpointAddress	EP1in
 0x03,			// bmAttributes		interrupt
 W(1<<EP1SSH),		// wMaxPacketSize
 1,			// bInterval		hier: aggresiv abfragen
#endif
#if WUB
 9,			// bLength
 4,			// bDescriptorType
 HID,			// bInterfaceNumber
 0,			// bAlternateSetting
 2-(HID&CDC),		// bNumEndpoints
 0xFF,			// bIfaceClass
 0,			// bIfaceSubClass
 0,			// bIfaceProtocol
 (4+HID)*(HID|CDC),	// iInterface
# if!(HID&CDC)
// endpoint descriptor
 7,			// bLength
 5,			// bDescriptorType
 CDC?0x81:0x83,		// bEndpointAddress	EP1in @ CDC sonst EP3in
 0x02,			// bmAttributes		bulk
 W(1<<(CDC?EP1SSH:EP3SSH)),// wMaxPacketSize
 0,			// bInterval
# endif
// endpoint descriptor
 7,			// bLength
 5,			// bDescriptorType
 CDC?0x02:0x04,		// bEndpointAddress	EP2out @ CDC sonst EP4out
 0x02,			// bmAttributes		bulk
 W(1<<(CDC?EP2SSH:EP4SSH)),// wMaxPacketSize
 0,			// bInterval
#endif
#if CDC
# if IAD
// interface assocoation descriptor (required for Windows >7 after HID)
 8,			// bLength
 11,			// bDescriptorType	iface assoc
 HID+WUB,		// bFirstIface
 2,			// bIfaceCount
 2,			// bFuncClass		CDC
 2,			// bFuncSubClass
 1,			// bFuncProtocol
 4+HID+WUB,		// iFunction
# endif
// interface descriptor
 9,			// bLength
 4,			// bDescriptorType	iface
 HID+WUB,		// bInterfaceNumber
 0,			// bAlternateSetting
 1,			// bNumEndpoints
 2,			// bIfaceClass		CDC Control
 2,			// bIfaceSubClass	abstract control model
 1,			// bIfaceProtocol	V.25ter
 (4+HID+WUB)*(1-IAD),	// iInterface
//HEADER_FUNCTIONAL_DESCRIPTOR
 5,			// bFunctionLength
 0x24,			// bDescriptorType	CS_INTERFACE
 0,			// bDescriptorSubtype	header functional descriptor
 W(0x0110),		// bcdCDC		specification version, 1.1
//ABSTRACT_CONTROL_MANAGEMENT_FUNCTIONAL_DESCRIPTOR
 4,			// bFunctionLength
 0x24,			// bDescriptorType	CS_INTERFACE
 2,			// bDescriptorSubtype	abstract control management functional descriptor
 2,			// bmCapabilities	nur SET_LINE_CODING, GET_LINE_CODING, SET_CONTROL_LINE_STATE, SERIAL_STATE
//UNION_FUNCTIONAL_DESCRIPTOR
 5,			// bFunctionLength
 0x24,			// bDescriptorType	CS_INTERFACE
 6,			// bDescriptorSubtype	union functional descriptor
 HID+WUB,		// bMasterInterface
 HID+WUB+1,		// bSlaveInterface0
//CALL_MANAGEMENT_FUNCTIONAL_DESCRIPTOR
 5,			// bFunctionLength
 0x24,			// bDescriptorType	CS_INTERFACE
 1,			// bDescriptorSubtype	call management functional descriptor
 3,			// bmCapabilities
 HID+WUB+1,		// bDataInterface
// dummy-endpoint descriptor
 7,			// bLength
 5,			// bDescriptorType
 WUB?0x85:0x82,		// bEndpointAddress
 0x03,			// bmAttributes		interrupt
 WUB?8:W(1<<EP2SSH),	// wMaxPacketSize
 WUB?255:8,		// bInterval
// interface descriptor
 9,			// bLength
 4,			// bDescriptorType
 HID+WUB+1,		// bInterfaceNumber
 0,			// bAlternateSetting
 2,			// bNumEndpoints
 10,			// bIfaceClass		CDC Data
 0,			// bIfaceSubClass
 0,			// bIfaceProtocol
 0,			// iInterface
// endpoint descriptor
 7,			// bLength
 5,			// bDescriptorType
 0x83,			// bEndpointAddress	3 in
 0x02,			// bmAttributes		bulk
 W(1<<EP3SSH),		// wMaxPacketSize	32
 0,			// bInterval
// endpoint descriptor
 7,			// bLength
 5,			// bDescriptorType
 0x04,			// bEndpointAddress	4 out
 0x02,			// bmAttributes		bulk
 W(1<<EP4SSH),		// wMaxPacketSize	16
 0,			// bInterval
#endif
};

template<byte L>struct StrDesc{
 char16_t wString[L>>1] {};
 constexpr StrDesc(const char16_t*s) {
  wString[0]= 3<<8|L;
  for (byte i=0; i<L-2>>1; i++) wString[i+1]=s[i];
 }
};

#define STRDESC(name,str) static const PROGMEM StrDesc<sizeof str> name(str)

STRDESC(string0,u"\u0407");	// 0407 = deutsch, 0409 = englisch, Mehrfachauswahl möglich
STRDESC(string1,u"Henrik Haftmann");
STRDESC(string2,u"Portalfräse");
STRDESC(string3,u"heha@hrz.tu-chemnitz.de:0031");

#if WUB

static const PROGMEM struct {
 uint16_t a,b;
 uint32_t c;
 uint16_t d;
# if HID|CDC
 uint16_t e,f,g,h,i,j,k,l;
# endif
 uint16_t m,n;
 char o[16];
 uint16_t p,q,r,s;
 char16_t t[21];
 uint16_t u;
 char16_t v[40];
}MsOs20Desc = {
// Microsoft OS 2.0 descriptor set header (table 10)
 10,	// Descriptor size (10 bytes)
 0,	// MS OS 2.0 descriptor set header
 0x06030000,	// Windows version (8.1) (0x06030000)
 sizeof MsOs20Desc,	// Size, MS OS 2.0 descriptor set
# if HID|CDC
// Microsoft OS 2.0 configuration subset header
 8,	// Descriptor size (8 bytes)
 1,	// MS OS 2.0 configuration subset header
 0,	// bConfigurationValue
 sizeof MsOs20Desc-10,	// Size, MS OS 2.0 configuration subset
// Microsoft OS 2.0 function subset header
 8,	// Descriptor size (8 bytes)
 2,	// MS OS 2.0 function subset header
 HID,	// interface number
 sizeof MsOs20Desc-18,	// Size, MS OS 2.0 function subset
# endif
// Microsoft OS 2.0 compatible ID descriptor (table 13)
 20,	// wLength
 3,	// MS_OS_20_FEATURE_COMPATIBLE_ID
 "WINUSB",
 4+21+1+40<<1,	//wLength: 
 4,	// wDescriptorType: MS_OS_20_FEATURE_REG_PROPERTY: 0x04 (Table 9)
 7,	// wPropertyDataType: REG_MULTI_SZ (Table 15)
 21<<1,	// wPropertyNameLength
 u"DeviceInterfaceGUIDs",	//bPropertyName
 40<<1,	// wPropertyDataLength
 u"{975F44D9-0D08-43FD-8B3E-127CA8AFFF9D}"	//bPropertyData
};

#define D(x) W(x),W((x)>>16)
static const PROGMEM byte BosDesc[5+24+28] = {
 5,		// Length
 0x0F,		// Binary Object Store descriptor
 W(sizeof BosDesc),		// Total length
 2,		// Number of device capabilities
// WebUSB Platform Capability descriptor
 24,		// Length
 0x10,		// Device Capability descriptor
 5,		// Platform Capability descriptor
 0,		// Reserved
 D(0x3408B638),W(0x09A9),W(0x47A0),0x8B,0xFD,0xA0,0x76,0x88,0x15,0xB6,0x65,  // WebUSB GUID
 W(0x0100),	// Version 1.0
 0x40,		// Vendor request code	(-> bRequest)
 0x01,		// Landing page		(-> wValueL)
// Microsoft OS 2.0 Platform Capability Descriptor
 28,		// Length
 0x10,		// Device Capability descriptor
 5,		// Platform Capability descriptor
 0,		// Reserved
 D(0xD8DD60DF),W(0x4589),W(0x4CC7),0x9C,0xD2,0x65,0x9D,0x9E,0x64,0x8A,0x9F,  // MS OS 2.0 GUID
 D(0x06030000),	// Windows version
 W(sizeof MsOs20Desc),	// Descriptor set length
 0x41,		// Vendor request code	(-> bRequest)
 0		// Alternate enumeration code
};
#undef D
static_assert(sizeof BosDesc==57,"Wrong struct!");

template<byte L>struct LandPageDesc{
 char a[L+3] {};
 constexpr LandPageDesc(char scheme,const char*s) {
  a[0]=L+3;
  a[1]=3;
  a[2]=scheme;
  for (byte i=0; i<L; i++) a[3+i]=s[i];
 }
};

#define LANDPAGE(name,scheme,str) static const PROGMEM LandPageDesc<sizeof str-1> name(scheme,str)

LANDPAGE(LandPage,1,u8"www.tu-chemnitz.de/~heha/ewa/Fräse/app.htm");
#endif

/**************************************************************************
 *  Variables - these are the only non-stack RAM usage
 **************************************************************************/

// zero when we are not configured, non-zero when enumerated
byte configval NOINIT;

cdc::DCB cdc::dcb = {1000000,0,0,8};

/**************************************************************************
 *  Public Functions - these are the API intended for the user
 **************************************************************************/

// initialize USB
void init() {
 USBCON = 1<<USBE | 1<<FRZCLK;	// enable USB
 PLLCSR = 1<<PLLE | 1<<PLLP0;	// config PLL
 while (!(PLLCSR & 1<<PLOCK));	// wait for PLL lock
 USBCON = 1<<USBE;		// start USB clock
 UDCON = 0;			// enable attach resistor
 configval = 0;
}

void poll() {
 byte intbits = UDINT;
 UDINT = 0;
 if (intbits & 1<<EORSTI) {
  UENUM = 0;
  UECONX = 1;
  UECFG0X = 0;	// control endpoint
  UECFG1X = (EP0SSH-3)<<4 | 2/*SINGLE_BUFFER*/;
  configval = 0;
 }
 if (intbits & 1<<SOFI) onSof();
 pollEp0();
}

// Misc functions to wait for ready and send/receive packets
static inline void usb_wait_in_ready() {
 while (!(UEINTX & 1<<TXINI));
}
static inline void usb_send_in() {
 UEINTX = ~(1<<TXINI);
}
static bool validEndpoint(byte ep) {
 return (HID && ep==0x81) || (CDC& (ep==0x82 || ep==0x83 || ep==0x04));
}

// Von-Neumannisiertes Speicherlesen (RAM oder Flash je nach Bit 15 der Adresse)
inline byte read_byte(const void*addr) {
 byte result;
 asm("	sbrc	%B1,7	\n"
 "	 lpm	%0,%a1	\n"
 "	sbrs	%B1,7	\n"
 "	 ld	%0,%a1	\n"
 :"=r"(result):"z"(addr));
 return result;
}

// Von-Neumannisierter Flash-Zeiger
inline const byte*FP(const void*x) {return (byte*)((unsigned)x+0x8000);}

// USB Endpoint Interrupt - endpoint 0 is handled here.  The
// other endpoints are manipulated by user-callable
// functions, and the start-of-frame interrupt.
void pollEp0() {
 UENUM = 0;
 byte intbits = UEINTX;
 if (intbits&1<<RXOUTI) {	// OUT angekommen?
  UEINTX=~(1<<RXOUTI);		// Daten ignorieren
 }
 if (intbits & 1<<RXSTPI) {
  byte bmRequestType = UEDATX,	// gesamtes SETUP-Paket in einzelne Register einlesen
	bRequest = UEDATX,
	wValueL = UEDATX,
	wValueH = UEDATX,
	wIndexL = UEDATX,
 UNUSED wIndexH = UEDATX,
	wLengthL = UEDATX,
	wLengthH = UEDATX,
	dlen;		// descriptor transfer length
  const byte*addr;	// descriptor address (in FLASH space)
  UEINTX=~(1<<RXSTPI);
// Fallstrick: Für EP0 müssen ERST die Daten ausgelesen werden
// und DANN das Interruptflag gelöscht werden.
// Letzteres setzt vmtl. TXINI und bereitet den Empfang von OUT-Daten vor.
  switch (bmRequestType) {
   case 0x80: switch (bRequest) {	// In,Std,Device
    case 6: switch (wValueH) {		// GetDescriptor((L=Index,H=DescType),LanguageID,Len)
     case 1: addr = FP(DeviceDesc);
gd1: dlen = read_byte(addr);
gd2: if (!wLengthH && dlen>wLengthL) dlen = wLengthL;
     byte n;	// needed outside loop for zero-length-packet detection!
     do{	// wait for host ready for IN packet
      byte i;
      do i = UEINTX; while (!(i & (1<<TXINI|1<<RXOUTI)));
      if (i & (1<<RXOUTI)) return;	// abort
      n = 1<<EP0SSH;			// send IN packet:
      if (n>dlen) n=dlen;		// it's a short packet, possibly zero!
      for (i = n; i; i--) UEDATX = read_byte(addr++);
      dlen -= n;
      usb_send_in();
     }while (dlen || n == 1<<EP0SSH);
     return;

     case 2: addr = FP(ConfigDesc); dlen = sizeof ConfigDesc; goto gd2;

     case 3: switch (wValueL) {
      case 0: addr = FP(&string0); goto gd1;
      case 1: addr = FP(&string1); goto gd1;
      case 2: addr = FP(&string2); goto gd1;
      case 3: addr = FP(&string3); goto gd1;
// Weitere String-Deskriptoren nur bei mehreren Interfaces
#if HID*(WUB|CDC)
      STRDESC(string4,u"HID-Interface");
      case 4: addr = FP(&string4); goto gd1;
#endif
#if WUB*(HID|CDC)
      STRDESC(string5,u"WebUSB");
      case 4+HID: addr = FP(&string5); goto gd1;
#endif
#if CDC
      STRDESC(string6,u"grbl und g-code");
      case 4+HID+WUB: addr = FP(&string6); goto gd1;
#endif
     }break;
#if WUB
     case 15: addr=FP(BosDesc); dlen=sizeof BosDesc; goto gd2;
#endif
#if HID
     case 0x21:
     case 0x22: goto usbtreeview;
#endif
    }break;

    case 8: {				// GetConfig(0,0,1)
     usb_wait_in_ready();
     UEDATX = configval;
     usb_send_in();
    }return;

    case 0: {				// GetStatus(0,0,2)
     UEDATX = 0;	// Bit0: Self-Power, Bit1: Remote Wakeup
     UEDATX = 0;
     usb_send_in();
    }return;
   }break;
   
   case 0x81: switch (bRequest) {	// In,Std,Iface
    case 0: {				// GetStatus(0,iface,2)
     UEDATX = 0;	// Stets 0
     UEDATX = 0;
     usb_send_in();
    }return;
    
#if HID
    case 6: switch (wIndexL) {		// GetDescriptor((L=?,H=?),iface,len)
     case 0: usbtreeview: switch (wValueH) {
      case 0x21:				// HidGetHidDesc(0x2100,iface,len)
      addr = FP(ConfigDesc)+HID_DESC_OFFSET; goto gd1;
      case 0x22: 		// HidGetReportDesc((L=descIndex,H=0x22),iface,len)
      addr = FP(HidRepDesc); dlen=sizeof HidRepDesc; goto gd2;
     }break;
    }break;
#endif

    case 10: {				// GetInterface(0,iface,1)
     UEDATX = 0;			// return one-and-only alternate setting 0
     usb_send_in();
    }return;
   }break;

   case 0x82: switch (bRequest) {	// In,Std,Endpoint
    case 0: {				// GetStatus(0,endpoint,2)
     if (!validEndpoint(wIndexL)) break;
     UENUM = wIndexL&7;
     byte answer = UECONX & 1<<STALLRQ ? 1 : 0;
     UENUM = 0;
     UEDATX = answer;	// Bit0: Endpoint stalled
     UEDATX = 0;
     usb_send_in();
    }return;
    case 18: {				// SyncFrame(0,endpoint,2)
     UEDATX = UDFNUML;
     UEDATX = UDFNUMH;
     usb_send_in();
    }break;
   }break;

   case 0x00: switch (bRequest) {	// Out,Std,Device
    case 1:				// ClearFeature(feature,0,0)
    case 3: return;			// SetFeature(feature,0,0)

    case 5: {				// SetAddr(addr,0,0)
     usb_send_in();
     usb_wait_in_ready();
     UDADDR = wValueL | 1<<ADDEN;	// Jetzt erst ändern!
    }return;

    case 9: {				// SetConfig(config,0,0)
     configval = wValueL;
     usb_send_in();
     for (byte i=0; i<4; i++) {
      UENUM = 1+i;
      UECONX = 1;
      static const byte cfg0[4] PROGMEM = {0xC1,0xC1,0x81,0x80};
      UECFG0X = pgm_read_byte(cfg0+i);	// Bulk/Interrupt In/Out
      static const byte cfg1[4] PROGMEM = {(EP1SSH-3)<<4|2,(EP2SSH-3)<<4|2,(EP3SSH-3)<<4|6,(EP4SSH-3)<<4|6};
      UECFG1X = pgm_read_byte(cfg1+i);	// size and single/double buffer
     }
     UERST = 0x1E;			// Togglebits rücksetzen
     UERST = 0;
    }return;
   }break;
   
   case 0x01: switch (bRequest) {	// Out,Std,Iface
    case 1:				// ClearFeature(sel,iface,0)
    case 3:				// SetFeature(sel,iface,0)
    case 11: return;			// SetInterface(alt_setting,iface,0)
   }break;

   case 0x02: switch (bRequest) {	// Out,Std,Endpoint
    case 1:				// ClearFeature(feature,endpoint,0)
    case 3: {				// SetFeature(feature,endpoint,0)
     if (wValueL) break;		// goto stall
     if (!validEndpoint(wIndexL)) break;
     usb_send_in();
     UENUM = wIndexL&=7;
     if (bRequest&2) UECONX = 1<<STALLRQ | 1<<EPEN;
     else{
      UECONX = 1<<STALLRQC | 1<<RSTDT | 1<<EPEN;
      UERST = 1<<wIndexL;		// Togglebit rücksetzen
      UERST = 0;
     }
    }return;
   }break;

   case 0x21: switch (wIndexL) {	// Out,Class,Iface
#if HID   
    case 0: switch (bRequest) {		// Iface=HID
     case 0x10: {			// HidSetReport((L=reportID,H=reportType),iface,len)
      hid::onSetReport(wValueH<<8|wValueL,wLengthH<<8|wLengthL);
     }break;
    }break;
#endif
#if CDC
    case HID+WUB: switch (bRequest) {	// Iface=CDC
     case 0x20: {			// CdcSetLineCoding(0,iface,7)
      if (wLengthL!=7) break;
      while (!(UEINTX & 1<<RXOUTI));
      for (byte i=0; i<7; i++) cdc::dcb[i]=UEDATX;
      UEINTX = ~(1<<RXOUTI);
      usb_send_in();
      cdc::onDcb();
     }return;
// Trotzdem kommt Linux „dmesg -w“ mit „cdc_acm failed to set dtr/rts“
// Bei Vendor-ID (VID) 0x2341 = 9025 = Arduino passiert das nicht!!
// Das bedeutet: Beim (generischen, vorinstallierten) Modem-Treiber
// ist diese VID fest eincompiliert!
// Wird diese VID verwendet, zieht sich der Modem-Treiber zurück.
// Ansonsten produziert dieser eine Menge Traffic auf dem Datenleitungen
// und versucht anscheinend, über AT-Kommandos RTS und DTR abzufragen.
// Sieht man bei mir mit „sudo cat /sys/kernel/debug/usb/usbmon/5u“.
// Weil Wireshark nicht gehen will.
// Um den Modemtreiber generell zu deaktivieren gibt es andere Wege.
     case 0x22: {			// CdcSetControlLineState(lineState,iface,0)
      if (wLengthL) break;
      cdc::onRtsDsr(wValueL);
      usb_wait_in_ready();
      usb_send_in();
      return;
     }
    }break;				// Bit 1 = RTS, Bit 0 = DTR
#endif
   }break;

   case 0xA1: switch (wIndexL) {	// In,Class,Iface
#if HID
    case 0: switch (bRequest) {		// Iface=HID
     case 1: {				// HidGetReport((L=reportID,H=reportType),iface,len)
      hid::onGetReport(wValueH<<8|wValueL,wLengthH<<8|wLengthL);
     }return;
    }break;
#endif
#if CDC
    case HID+WUB: switch (bRequest) {
     case 0x20: {	// CdcGetSerialState(0,iface,2)
      UEDATX = 0;	// Bits: 6 = Overrun, 5 = Parity, 4 = Framing
      UEDATX = 0;	// 3 = RI, 2 = Break, 1 = DSR, 0 = DCD
      usb_send_in();
     }return;
     case 0x21: addr = cdc::dcb; dlen = sizeof cdc::dcb; goto gd2;	// CdcGetLineCoding(0,iface,7)
    }break;
#endif
   }break;

#if WUB
   case 0xC0: switch (bRequest) {	// In, Vendor, Device
    case 0x40: addr=FP(&LandPage); goto gd1;	// wValue=1, wIndex=2
    case 0x41: addr=FP(&MsOs20Desc); dlen=sizeof MsOs20Desc; goto gd2;	// wValue=0, wIndex=7
   }break;
#endif

  }
  UECONX = 1<<STALLRQ | 1<<EPEN;	// stall EP0 on switch-fallthrough or "break"
 }
}

}//namespace
